#include "PlyLoader.h"
#include "AdaptiveMesh.h"
#include <string.h>
#include <cstring>
#include <cstdlib>

#ifdef _MSC_VER
  #define strncasecmp strnicmp
  #define strcasecmp stricmp
#endif


void PlyLoader::triangulate ( char **faces, int numFaces, AdaptiveArray<char *> *faceArray )
{
}
	

void getToken ( FILE *fptr, char *token )
{
	int index = 0;
	char temp;

	// ignore leading white space
	fread ( &temp, 1, 1, fptr );
	while ( temp == ' ' || temp == '\n' || temp == '\r' || temp == '\t' )
	{
		fread ( &temp, 1, 1, fptr );
	}

	while ( temp != ' ' && temp != '\n' && temp != '\r' && temp != '\t' )
	{
		token [ index++ ] = temp;
		fread ( &temp, 1, 1, fptr );
	}

	token [ index ] = '\0';
}

void readUntilReturn ( FILE *fptr )
{
	char temp;

	do
	{
		fread ( &temp, 1, 1, fptr );
	} while ( temp != '\n' || '\r' );
}

void readLine ( FILE *fptr, char *buffer )
{
	char temp;
	int index = 0;

	fread ( &temp, 1, 1, fptr );
	while ( temp != '\n' && temp != '\r' )
	{
		buffer [ index++ ] = temp;
		fread ( &temp, 1, 1, fptr );
	}

	buffer [ index ] = '\0';

	// if we read a comment line, just ignore it and read the next line
	if ( strncasecmp ( buffer, "comment", 7 ) == 0 || strlen ( buffer ) == 0 )
	{
		readLine ( fptr, buffer );
	}
}

void flipBits16 ( void *x )
{
	unsigned char *temp = (unsigned char *)x;
	unsigned char swap;
	
	swap = temp [ 0 ];
	temp [ 0 ] = temp [ 1 ];
	temp [ 1 ] = swap;
}

void flipBits32 ( void *x )
{
	unsigned char *temp = (unsigned char *)x;
	unsigned char swap;
	
	swap = temp [ 0 ];
	temp [ 0 ] = temp [ 3 ];
	temp [ 3 ] = swap;

	swap = temp [ 1 ];
	temp [ 1 ] = temp [ 2 ];
	temp [ 2 ] = swap;
}

Mesh* PlyLoader::loadFile ( char *filename )
{
	AdaptiveMesh *rvalue = NULL;
	FILE *fptr;
	char buffer [ 1024 ];
	char *tempPtr;
	bool ascii = true;
	bool big_endian = true;
	int sizeOfVertex = 0;
	int sizeOfFaceElement = 0;
	int sizeOfFaceNum = 0;
	int numFaceElements = 0;
	int numFaces = 0;
	int numVertices = 0;
	int i, j;
  int numTriStrips = 0;

	fptr = fopen ( filename, "rb" );
	if ( fptr == NULL )
	{
		fprintf ( stderr, "Failed to load model: %s\n", filename );
		exit ( 1 );
	}

	getToken ( fptr, buffer );

	if ( strcasecmp ( buffer, "ply" ) != 0 )
	{
		fprintf ( stderr, "ERROR: This is not a ply file!\n" );
		exit ( 1 );
	}

	getToken ( fptr, buffer );
	if ( strcasecmp ( buffer, "format" ) != 0 )
	{
		fprintf ( stderr, "ERROR: Expected 'format' and read %s\n", buffer );
		exit ( 1 );
	}

	getToken ( fptr, buffer );
	if ( strcasecmp ( buffer, "ascii" ) == 0 )
	{
		ascii = true;
	}
	else
	{
		ascii = false;

		if ( strcasecmp ( buffer, "binary_big_endian" ) == 0 )
		{
			big_endian = true;
		}
		else
		{
			big_endian = false;
		}
	}

	// stupid version number
	getToken ( fptr, buffer );

	// begin reading data definition lines
	readLine ( fptr, buffer );
	if ( strncasecmp ( buffer, "element vertex", strlen ( "element vertex" ) ) != 0 )
	{
		fprintf ( stderr, "ERROR: Expected 'element vertex' but read %s\n", buffer );
		exit ( 1 );
	}
	strtok ( buffer, " " );
	strtok ( NULL, " " );
	tempPtr = strtok ( NULL, " " );

	numVertices = atoi ( tempPtr );

	readLine ( fptr, buffer );
	while ( strncasecmp ( buffer, "element", 7 ) != 0 )
	{
		if ( strncasecmp ( buffer, "property", 8 ) != 0 )
		{
			fprintf ( stderr, "ERROR: Expected 'property' but read %s\n", buffer );
			exit ( 1 );
		}

		strtok ( buffer, " " );
		tempPtr = strtok ( NULL, " " );

		if ( strcasecmp ( tempPtr, "char" ) == 0 || strcasecmp ( tempPtr, "uchar" ) == 0 ||
			strcasecmp ( tempPtr, "uint8" ) == 0 )
		{
			sizeOfVertex += 1;
		} else if ( strcasecmp ( tempPtr, "short" ) == 0 || strcasecmp ( tempPtr, "ushort" ) == 0 )
		{
			sizeOfVertex += 2;
		} else if ( strcasecmp ( tempPtr, "int" ) == 0 || strcasecmp ( tempPtr, "uint" ) == 0 || 
			strcasecmp ( tempPtr, "float" ) == 0 || strcasecmp ( tempPtr, "float32" ) == 0 ||
			strcasecmp ( tempPtr, "int32" ) == 0 )
		{
			sizeOfVertex += 4;
		} else if ( strcasecmp ( tempPtr, "double" ) == 0 )
		{
			sizeOfVertex += 8;
		} else
		{
			fprintf ( stderr, "ERROR: unknown data type: %s\n", tempPtr );
			exit ( 1 );
		}

		readLine ( fptr, buffer );
	}

	if ( strncasecmp ( buffer, "element face", 12 ) != 0 )
	{
    if ( strncasecmp ( buffer, "element tristrip", 16 ) != 0 )
    {
		  fprintf ( stderr, "ERROR: Expected 'element face' but read %s\n", buffer );
		  exit ( 1 );
    }
	  strtok ( buffer, " " );
	  strtok ( NULL, " " );
	  tempPtr = strtok ( NULL, " " );

	  numTriStrips = atoi ( tempPtr );    
	}
  else
  {
	  strtok ( buffer, " " );
	  strtok ( NULL, " " );
	  tempPtr = strtok ( NULL, " " );

	  numFaces = atoi ( tempPtr );
  }

	readLine ( fptr, buffer );
	if ( strncasecmp ( buffer, "property list", 13 ) != 0 )
	{
		fprintf ( stderr, "ERROR: Expected 'property list' but read %s\n", buffer );
		exit ( 1 );
	}
	strtok ( buffer, " " );
	strtok ( NULL, " " );
	tempPtr = strtok ( NULL, " " );

	if ( strcasecmp ( tempPtr, "char" ) == 0 || strcasecmp ( tempPtr, "uchar" ) == 0 ||
		strcasecmp ( tempPtr, "uint8" ) == 0 )
	{
		sizeOfFaceNum = 1;
	} else if ( strcasecmp ( tempPtr, "short" ) == 0 || strcasecmp ( tempPtr, "ushort" ) == 0 )
	{
		sizeOfFaceNum = 2;
	} else if ( strcasecmp ( tempPtr, "int" ) == 0 || strcasecmp ( tempPtr, "uint" ) == 0 ||
		strcasecmp ( tempPtr, "int32" ) == 0 )
	{
		sizeOfFaceNum = 4;
	} else
	{
		fprintf ( stderr, "ERROR: unknown data type for size of face list: %s\n", tempPtr );
		exit ( 1 );
	}

	tempPtr = strtok ( NULL, " " );
	if ( strcasecmp ( tempPtr, "char" ) == 0 || strcasecmp ( tempPtr, "uchar" ) == 0 ||
		strcasecmp ( tempPtr, "uint8" ) == 0 )
	{
		sizeOfFaceElement = 1;
	} else if ( strcasecmp ( tempPtr, "short" ) == 0 || strcasecmp ( tempPtr, "ushort" ) == 0 )
	{
		sizeOfFaceElement = 2;
	} else if ( strcasecmp ( tempPtr, "int" ) == 0 || strcasecmp ( tempPtr, "uint" ) == 0 ||
		strcasecmp ( tempPtr, "int32" ) == 0 )
	{
		sizeOfFaceElement = 4;
	} else
	{
		fprintf ( stderr, "ERROR: unknown data type for face index: %s\n", tempPtr );
		exit ( 1 );
	}

	// ignore everything else until end of header
	readLine ( fptr, buffer );
	while ( strncasecmp ( buffer, "end_header", 10 ) != 0 )
	{
		readLine ( fptr, buffer );
	}
/*
	if ( sizeOfFaceNum != 1 )
	{
		fprintf ( stderr, "ERROR: element for number of sides of a faces is bigger than 1 byte!\n" );
		exit ( 1 );
	}
*/
	rvalue = new AdaptiveMesh ( 3 );

	Vertex vertex ( 3 );
	Face face ( 3 );
	int n;
	float fbuffer [ 30 ];

	if ( ascii )
	{
		for ( i = 0; i < numVertices; i++ )
		{
			readLine ( fptr, buffer );
			vertex.setX ( atof ( strtok ( buffer, " " ) ) );
			vertex.setY ( atof ( strtok ( NULL, " " ) ) );
			vertex.setZ ( atof ( strtok ( NULL, " " ) ) );

//			printf ( "Added vertex %d: %f,%f,%f\n", i, vertex [ 0 ], vertex [ 1 ], vertex [ 2 ] );
			rvalue->addVertex ( &vertex );
		}

		for ( i = 0; i < numFaces; i++ )
		{
			readLine ( fptr, buffer );
			n = atoi ( strtok ( buffer, " " ) );
			int *indices = new int [ n ];

			if ( n > 3 )
			{
				fprintf ( stderr, "ERROR: can't read a face of more than 3 sides (file had %d)\n", n );
				exit ( 1 );
			}

			for ( j = 0; j < n; j++ )
			{
				indices [ j ] = atoi ( strtok ( NULL, " " ) );
			}

			// triangulate the faces
			face.getData ( ) [ 0 ] = indices [ 0 ];
			for ( j = 1; j < n - 1; j++ )
			{
				face.getData ( ) [ 1 ] = indices [ j ];
				face.getData ( ) [ 2 ] = indices [ j +  1 ];
				rvalue->addFace ( &face );
			}

			delete[] indices;
		}
	}
	else
	{
		if ( !big_endian )
		{
			for ( i = 0; i < numVertices; i++ )
			{
				fread ( fbuffer, 1, sizeOfVertex, fptr );

				vertex.setX ( fbuffer [ 0 ] );
				vertex.setY ( fbuffer [ 1 ] );
				vertex.setZ ( fbuffer [ 2 ] );

				rvalue->addVertex ( &vertex );
			}

			for ( i = 0; i < numFaces; i++ )
			{
				fread ( buffer, 1, sizeOfFaceNum, fptr );

				n = buffer [ 0 ];
				int *indices = new int [ n ];

				for ( j = 0; j < n; j++ )
				{
					fread ( &(indices [ j ]), 1, sizeOfFaceElement, fptr );
				}

				// triangulate the faces
				face.getData ( ) [ 0 ] = indices [ 0 ];
				for ( j = 1; j < n - 1; j++ )
				{
					face.getData ( ) [ 1 ] = indices [ j ];
					face.getData ( ) [ 2 ] = indices [ j +  1 ];
					rvalue->addFace ( &face );
				}

				delete[] indices;
			}

			for ( i = 0; i < numTriStrips; i++ )
			{
				fread ( buffer, 1, sizeOfFaceNum, fptr );

				n = *((int *)buffer);
				int *indices = new int [ n ];

				for ( j = 0; j < n; j++ )
				{
					fread ( &(indices [ j ]), 1, sizeOfFaceElement, fptr );
				}

				face.getData ( ) [ 0 ] = indices [ 0 ];
        bool flipOrder = false;
				for ( j = 1; j < n - 1; j++ )
				{
          if ( indices [ j + 1 ] == -1 )
          {
            flipOrder = false;
            j += 2;
          }
          else
          {
            if ( !flipOrder )
            {
              face.getData ( ) [ 0 ] = indices [ j - 1 ];
              face.getData ( ) [ 1 ] = indices [ j ];
              face.getData ( ) [ 2 ] = indices [ j + 1 ];
            }
            else
            {
              face.getData ( ) [ 0 ] = indices [ j + 1 ];
              face.getData ( ) [ 1 ] = indices [ j ];
              face.getData ( ) [ 2 ] = indices [ j - 1 ];
            }
            flipOrder = !flipOrder;
					  rvalue->addFace ( &face );
          }
				}

				delete[] indices;
			}
		}
		else
		{
			for ( i = 0; i < numVertices; i++ )
			{
				fread ( fbuffer, 1, sizeOfVertex, fptr );

				flipBits32 ( &fbuffer [ 0 ] );
				flipBits32 ( &fbuffer [ 1 ] );
				flipBits32 ( &fbuffer [ 2 ] );

				vertex.setX ( fbuffer [ 0 ] );
				vertex.setY ( fbuffer [ 1 ] );
				vertex.setZ ( fbuffer [ 2 ] );
				rvalue->addVertex ( &vertex );
			}

			for ( i = 0; i < numFaces; i++ )
			{
				fread ( buffer, 1, sizeOfFaceNum, fptr );

				n = buffer [ 0 ];
				int *indices = new int [ n ];

				if ( sizeOfFaceElement == 2 )
				{
					for ( j = 0; j < n; j++ )
					{
						fread ( &(indices [ j ]), 1, sizeOfFaceElement, fptr );
						flipBits16 ( &(indices [ j ]) );
					}
				}
				else
				{
					for ( j = 0; j < n; j++ )
					{
						fread ( &(indices [ j ]), 1, sizeOfFaceElement, fptr );
						flipBits32 ( &(indices [ j ]) );
					}
				}

				// triangulate the faces
				face.getData ( ) [ 0 ] = indices [ 0 ];
				for ( j = 1; j < n - 1; j++ )
				{
					face.getData ( ) [ 1 ] = indices [ j ];
					face.getData ( ) [ 2 ] = indices [ j +  1 ];
					rvalue->addFace ( &face );
				}

				delete[] indices;
			}

			for ( i = 0; i < numTriStrips; i++ )
			{
				fread ( buffer, 1, sizeOfFaceNum, fptr );

				n = *((int *)buffer);
        flipBits32 ( &n );
				int *indices = new int [ n ];

				if ( sizeOfFaceElement == 2 )
				{
					for ( j = 0; j < n; j++ )
					{
						fread ( &(indices [ j ]), 1, sizeOfFaceElement, fptr );
						flipBits16 ( &(indices [ j ]) );
					}
				}
				else
				{
					for ( j = 0; j < n; j++ )
					{
						fread ( &(indices [ j ]), 1, sizeOfFaceElement, fptr );
						flipBits32 ( &(indices [ j ]) );
					}
				}

				face.getData ( ) [ 0 ] = indices [ 0 ];
        bool flipOrder = false;
				for ( j = 1; j < n - 1; j++ )
				{
          if ( indices [ j + 1 ] == -1 )
          {
            flipOrder = false;
            j += 2;
          }
          else
          {
            if ( !flipOrder )
            {
              face.getData ( ) [ 0 ] = indices [ j - 1 ];
              face.getData ( ) [ 1 ] = indices [ j ];
              face.getData ( ) [ 2 ] = indices [ j + 1 ];
            }
            else
            {
              face.getData ( ) [ 0 ] = indices [ j + 1 ];
              face.getData ( ) [ 1 ] = indices [ j ];
              face.getData ( ) [ 2 ] = indices [ j - 1 ];
            }
            flipOrder = !flipOrder;
					  rvalue->addFace ( &face );
          }
				}

				delete[] indices;
			}
		}
	}

	fclose ( fptr );
	
	// print info out
#ifdef _DEBUG
	printf ( "Type of file " );
	if ( ascii )
	{
		printf ( "ascii\n" );
	} else if ( big_endian )
	{
		printf ( "binary_big_endian\n" );
	} else
	{
		printf ( "binary_little_endian\n" );
	}

	printf ( "Number of vertices %d\n", rvalue->getNumVertices ( ) );
	printf ( "Size of vertex %d\n", sizeOfVertex );
	printf ( "Number of faces %d\n", rvalue->getNumFaces ( ) );
	printf ( "Size of num face elements %d\n", sizeOfFaceNum );
	printf ( "Size of face index %d\n", sizeOfFaceElement );
#endif

	return rvalue;
}
